"""Wrapper around CSS vector database."""
from __future__ import annotations

import logging
import uuid
from abc import ABC
from typing import Any, Dict, List, Optional, Tuple, Callable

from langchain.embeddings.base import Embeddings
from langchain.utils import get_from_env

from langchain_pangu.pangukitsappdev.api.memory.vector.base import VectorStore, Document
from langchain_pangu.pangukitsappdev.utils.time_date import now_millis
from langchain_pangu.pangukitsappdev.vectorstores.bulk_data import BulkData

logger = logging.getLogger(__name__)


def _default_text_mapping(dim: int, metric: str = "inner_product") -> Dict:
    return {
        "mappings": {
            "properties": {
                "text": {"type": "text"},
                "vector": {
                    "type": "vector",
                    "dimension": dim,
                    "indexing": 'true',
                    "algorithm": "GRAPH",
                    "metric": metric
                },
            }
        }, "settings": {"index": {"vector": "true"}}}


def _default_script_query(query_vector: List[float], vector_field: str = "vector",
                          metric: str = "inner_product") -> Dict:
    return {
        "script_score": {
            "query": {"match_all": {}},
            "script": {
                "source": "vector_score",
                "lang": "vector",
                "params": {
                    "field": vector_field,
                    "vector": query_vector,
                    "metric": metric
                }
            }
        }
    }


def _default_get_doc_with_score_func(hit: dict, text_key: str) -> Tuple[Document, float]:
    return (
        Document(
            page_content=str(hit["_source"].get(text_key)),
            metadata=hit["_source"],
            id=hit["_id"],
            score=hit["_score"]
        ),
        hit["_score"],
    )


# CSSVectorSearch是抽象基类VectorStore的具体实现，它为所有矢量数据库实现定义了一个公共接口。
# 通过继承ABC类，弹性矢量搜索可以定义为抽象基类本身，允许创建具有自己特定实现的子类。
class CSSVectorSearch(VectorStore, ABC):
    """把CSS数据库包装成向量数据库，CSS是基于开源Elasticsearch开发的自研向量搜索数据库，连接不需要身份认证信息, 传入 URL 和 index 以及
    embedding 进行构造.
    Args:
        elasticsearch_url (str): The URL for the Elasticsearch instance.
        index_name (str): The name of the Elasticsearch index for the embeddings.
        embedding (Embeddings): An object that provides the ability to embed text.
                It should be an instance of a class that subclasses the Embeddings
                abstract base class, such as OpenAIEmbeddings()

    Raises:
        ValueError: If the elasticsearch python package is not installed.
    """

    def __init__(
            self,
            elasticsearch_url: str,
            index_name: str,
            embedding: Embeddings,
            vector_fields: List[str],
            source_fields: List[str],
            text_key: str,
            mapping_func: Callable[[int], Dict] = _default_text_mapping,
            script_query_func: Callable[[List[float], str], dict] = _default_script_query,
            get_doc_with_score_func: Callable[[Dict, str], Tuple[Document, float]] = _default_get_doc_with_score_func,
            **kwargs
    ):
        """
        初始化
        Args:
            elasticsearch_url: 作为elasticsearch.client.Elasticsearch的hosts参数，支持逗号分割的list
            index_name: 索引名称
            embedding: （Optional）embedding接口的实现类。如果不传递，则走文本检索的逻辑
            mapping_func: （Optional）支持传递自定义的构造mapping函数，允许自定义mapping。入参为向量字段的长度。返回索引的mapping
            script_query_func: （Optional）支持传递自定义的查询dsl构造方法，入参为待搜索的向量vector: List[float]。返回query dsl
            get_doc_with_score_func: (Optional)传递自定义的从检索结果获取（Document, score）数据的函数，入参是es的检索结果
            add_text_request_func: (Optional)传递自定义的函数，在批量添加索引数据时构造请求体。入参：索引名，向量，文档内容，文档元数据\n
                出参（文档id, 构造的请求体）
            **kwargs:
        """

        """Initialize with necessary components."""
        try:
            import elasticsearch
        except ImportError:
            raise ImportError(
                "Could not import elasticsearch python package. "
                "Please install it with `pip install elasticsearch`."
            )
        self.embedding = embedding
        self.index_name = index_name
        self.mapping_func = mapping_func
        self.script_query_func = script_query_func
        self.get_doc_with_score_func = get_doc_with_score_func
        self.vector_fields = vector_fields
        self.source_fields = source_fields
        self.text_key = text_key
        try:
            from langchain_pangu.pangukitsappdev.vectorstores.proxy_http_requests import ProxyRequestsHttpConnection
            self.client = elasticsearch.Elasticsearch(elasticsearch_url, connection_class=ProxyRequestsHttpConnection,
                                                      **kwargs)
        except ValueError as e:
            raise ValueError(
                f"Your elasticsearch client string is mis-formatted. Got error: {e} "
            )

    def add_docs(self, bulk_data_list: List[BulkData], **kwargs):
        batch_size = kwargs.get("batch_size", 100)
        # 一批一批的数据进行索引构建
        sub_bulk_data = []
        for i, d in enumerate(bulk_data_list):
            sub_bulk_data.append(d)
            if (i + 1) % batch_size == 0:
                if self.is_external_embedding():
                    vector_field = self.vector_fields[0]
                    texts = [bulk.data.get(vector_field, "") for bulk in sub_bulk_data]
                    self.add_docs_with_external_embedding(sub_bulk_data,
                                                          self.embedding.embed_documents(texts))
                else:
                    self.add_docs_with_builtin_embedding(sub_bulk_data)
                sub_bulk_data = []
                logger.info("Indexed %s document", str(i+1))

        if sub_bulk_data:
            if self.is_external_embedding():
                vector_field = self.vector_fields[0]
                texts = [bulk.data.get(vector_field, "") for bulk in sub_bulk_data]
                self.add_docs_with_external_embedding(sub_bulk_data,
                                                      self.embedding.embed_documents(texts))
            else:
                self.add_docs_with_builtin_embedding(sub_bulk_data)

    def add_docs_with_external_embedding(self, bulk_data_list: List[BulkData], embedded: List[List[float]]):
        vector_field = self.vector_fields[0]
        bulk_requests = []

        if not self.client.indices.exists(index=self.index_name):
            # 外部embedding模式可自检索引
            if self.mapping_func:
                mappings = self.mapping_func(len(embedded[0]))
                if not self.client.indices.exists(index=self.index_name):
                    # 创建索引和映射
                    self.client.indices.create(index=self.index_name,
                                               body=mappings)
                    logger.info("Index created successfully.")
            else:
                raise ImportError(
                    "index is not exists. Indexes should be created first."
                )
        for i, bulk_data in enumerate(bulk_data_list):
            req = {
                "_op_type": "index",
                "_index": self.index_name,
                "_id": bulk_data.id,
            }
            req.update(bulk_data.data)
            req.update({vector_field: embedded[i]})
            bulk_requests.append(req)
        try:
            from elasticsearch.exceptions import NotFoundError
            from elasticsearch.helpers import bulk
        except ImportError:
            raise ImportError(
                "Could not import elasticsearch python package. "
                "Please install it with `pip install elasticsearch`."
            )
        bulk(self.client, bulk_requests)
        self.client.indices.refresh(index=self.index_name)

    def add_docs_with_builtin_embedding(self, bulk_data_list: List[BulkData]):
        bulk_requests = []
        if not self.client.indices.exists(index=self.index_name):
            # CSS add_docs 索引需外部创建好
            raise ImportError(
                "index is not exists. Indexes should be created first."
            )
        for i, bulk_data in enumerate(bulk_data_list):
            req = {
                "_op_type": "index",
                "_index": self.index_name,
                "_id": bulk_data.id,
            }
            req.update(bulk_data.data)
            bulk_requests.append(req)
        try:
            from elasticsearch.exceptions import NotFoundError
            from elasticsearch.helpers import bulk
        except ImportError:
            raise ImportError(
                "Could not import elasticsearch python package. "
                "Please install it with `pip install elasticsearch`."
            )
        bulk(self.client, bulk_requests)
        self.client.indices.refresh(index=self.index_name)

    def similarity_search(
            self, query: str, k: int = 5,
            score_threshold: float = -1.0, **kwargs: Any
    ) -> List[Document]:
        """返回和查询语句最相近的k条文本.

        Args:
            query: Text to look up documents similar to.
            k: Number of Documents to return. Defaults to 5.
            score_threshold: 低于这个阈值分数的doc不会被检索出来
        Returns:
            List of Documents most similar to the query.
        """
        docs_and_scores = self.similarity_search_with_score(query, k, score_threshold, **kwargs)
        documents = [d[0] for d in docs_and_scores]
        return documents

    def similarity_search_with_score(
            self, query: str, k: int = 5,
            score_threshold: float = -1.0,
            **kwargs: Any
    ) -> List[Tuple[Document, float]]:
        """Return docs most similar to query.
        Args:
            query: Text to look up documents similar to.
            k: Number of Documents to return. Defaults to 5.
            score_threshold: （Optional）低于这个阈值分数的doc不会被检索出来
        Returns:
            List of Documents most similar to the query.
        """

        if self.is_external_embedding():
            embedding = self.embedding.embed_query(query)
            docs_with_score = self.similarity_search_with_external_embedding(embedding, k, **kwargs)
        else:
            # 如果没有传递embedding，则执行如下逻辑
            docs_with_score = self.similarity_search_with_builtin_embedding(query=query, k=k, **kwargs)

        return [dws for dws in docs_with_score if dws[1] >= score_threshold]

    def similarity_search_with_external_embedding(self, embedding, k, **kwargs) -> List[Tuple[Document, float]]:
        vector_field = self.vector_fields[0]
        response = self.client.search(index=self.index_name,
                                      body={
                                          'query': self.script_query_func(embedding, vector_field),
                                          '_source': self.source_fields,
                                          'size': k
                                      }, size=k, **kwargs)
        hits = [hit for hit in response["hits"]["hits"]]
        docs_and_scores = [
            self.get_doc_with_score_func(hit, self.text_key)
            for hit in hits
        ]
        return docs_and_scores

    def similarity_search_with_builtin_embedding(self, query, k, **kwargs) -> List[Tuple[Document, float]]:
        """检索,不使用
        :param query: query词
        :param k: 检索结果数
        :param kwargs: 扩展参数
        :return: :class:`List[Document] <List[Document]>` 检索结果
        """

        query_dsl = {
            "query": {
                "multi_match": {
                    "query": query,
                    "fields": self.vector_fields
                }
            },
            "from": 0,
            "size": k,
            "_source": self.source_fields
        }
        start_millis = now_millis()
        response = self.client.search(index=self.index_name, body=query_dsl, size=k, **kwargs)
        end_millis = now_millis()

        duration_ms = end_millis - start_millis

        # 服务端处理的耗时信息
        server_cost_time = response.get("timestamp")
        # 服务端处理的状态
        server_status = response.get("status")

        logging.info(
            "Success return, request cost: [duration_ms: %s], [server_cost_time: %s], "
            "[server_status: %s]", str(duration_ms), str(server_cost_time), str(server_status))

        hits = [hit for hit in response["hits"]["hits"]]
        # 解析搜索结果，支持从参数中传递解析函数
        docs_and_scores = [
            self.get_doc_with_score_func(hit, self.text_key)
            for hit in hits
        ]
        return docs_and_scores

    def similarity_search_with_relevance_scores(self, query: str, k: int = 5,
                                                score_threshold: float = -1.0,
                                                **kwargs: Any) -> List[Tuple[Document, float]]:
        return self.similarity_search_with_score(query, k, score_threshold, **kwargs)

    def is_external_embedding(self) -> bool:
        if not self.embedding:
            return False
        if len(self.vector_fields) != 1:
            logger.error("add docs with external embedding, "
                         "but the vector field is not config or have multi vector fields")
            raise ValueError("add docs with external embedding, "
                             "but the vector field is not config or have multi vector fields")
        return True

    def dsl_search(self, dsl: dict) -> dict:
        response = self.client.search(index=self.index_name, body=dsl)
        return response

    def delete(self, ids: Optional[List[str]] = None, **kwargs: Any):
        delete_query = {
            "query": {
                "terms": {
                    "_id": ids
                }
            }
        }
        self.client.delete_by_query(index=self.index_name, body=delete_query)

    def clear(self):
        delete_query = {
            "query": {
                "match_all": {}
            }
        }
        self.client.delete_by_query(index=self.index_name, body=delete_query)

    @classmethod
    def from_texts(
            cls,
            texts: List[str],
            embedding: Embeddings,
            metadatas: Optional[List[dict]] = None,
            elasticsearch_url: Optional[str] = None,
            index_name: Optional[str] = None,
            refresh_indices: bool = True,
            **kwargs: Any,
    ) -> CSSVectorSearch:
        """Construct ElasticVectorSearch wrapper from raw documents.

        This is a user-friendly interface that:
            1. Embeds documents.
            2. Creates a new index for the embeddings in the Elasticsearch instance.
            3. Adds the documents to the newly created Elasticsearch index.

        This is intended to be a quick way to get started.

        Example:
            .. code-block:: python

                from langchain import ElasticVectorSearch
                from langchain.embeddings import OpenAIEmbeddings
                embeddings = OpenAIEmbeddings()
                elastic_vector_search = ElasticVectorSearch.from_texts(
                    texts,
                    embeddings,
                    elasticsearch_url="http://localhost:9200"
                )
        """
        elasticsearch_url = elasticsearch_url or get_from_env(
            "elasticsearch_url", "ELASTICSEARCH_URL"
        )
        index_name = index_name or uuid.uuid4().hex
        vector_search = cls(elasticsearch_url, index_name, embedding, **kwargs)
        vector_search.add_texts(
            texts, metadatas=metadatas, refresh_indices=refresh_indices
        )
        return vector_search



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from pydantic.v1 import BaseModel


class LLMResp(BaseModel):
    """
    answer: 回答
    is_from_cache: 是否命中缓存
    """

    answer: str
    is_from_cache: bool = False



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import json
import logging
from json import JSONDecodeError
from typing import List, Optional, Iterator, Any

import requests
import sseclient
from langchain.callbacks.manager import CallbackManagerForLLMRun
from langchain.chat_models.base import BaseChatModel
from langchain.llms.base import LLM, BaseLLM
from langchain.schema import AIMessage, BaseMessage
from langchain.schema.messages import AIMessageChunk
from langchain.schema.output import GenerationChunk, LLMResult, Generation, ChatGeneration, ChatResult, \
    ChatGenerationChunk
from requests.exceptions import ChunkedEncodingError

from langchain_pangu.pangukitsappdev.api.common_config import AUTH_TOKEN_HEADER
from langchain_pangu.pangukitsappdev.api.llms.base import AbstractLLMApi, get_llm_params, convert_message_to_req
from langchain_pangu.pangukitsappdev.api.llms.llm_config import LLMConfig
from langchain_pangu.pangukitsappdev.auth.iam import IAMTokenProvider, IAMTokenProviderFactory
from langchain_pangu.pangukitsappdev.llms.response.gallery_text_resp import GalleryUsage, GalleryTextResp, \
    GalleryTextChoice
from langchain_pangu.pangukitsappdev.llms.response.llm_response_gallery import LLMRespGallery

logger = logging.getLogger(__name__)


class GalleryLLM(LLM):
    temperature: Optional[float]
    max_tokens: Optional[int]
    top_p: Optional[float]
    gallery_url: str
    token_getter: IAMTokenProvider
    streaming: Optional[bool]
    proxies: dict = {}

    def _call(self, prompt: str, stop: Optional[List[str]] = None,
              run_manager: Optional[CallbackManagerForLLMRun] = None) -> str:
        llm_result = self._generate([prompt], stop, run_manager)
        return llm_result.generations[0][0].text

    @property
    def _llm_type(self) -> str:
        return "gallery_llm"

    def _stream(
            self,
            prompt: str,
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> Iterator[GenerationChunk]:
        request_body = {
            "prompt": prompt,
            "stream": True,
            **get_llm_params({
                "temperature": self.temperature,
                "max_tokens": self.max_tokens,
                "top_p": self.top_p
            })
        }
        token = self.token_getter.get_valid_token()
        headers = {
            AUTH_TOKEN_HEADER: token,
            "X-Agent": "pangu-kits-app-dev"
        } if token else {"X-Agent": "pangu-kits-app-dev"}
        rsp = requests.post(self.gallery_url, headers=headers,
                            json=request_body,
                            verify=False, stream=True, proxies=self.proxies)
        try:
            rsp.raise_for_status()
            stream_client: sseclient.SSEClient = sseclient.SSEClient(rsp)
            for event in stream_client.events():
                # 解析出Token数据
                data_json = json.loads(event.data)
                if data_json.get("choices") is None:
                    raise ValueError(f"Meet json decode error: {str(data_json)}, not get choices")
                chunk = GenerationChunk(text=data_json["choices"][0]["text"])
                yield chunk
                if run_manager:
                    run_manager.on_llm_new_token(chunk.text)
        except JSONDecodeError as ex:
            # [DONE]表示stream结束了
            if event.data != "[DONE]":
                logger.warning("Meet json decode error: %s", str(ex))
        except ChunkedEncodingError as ex:
            logger.warning("Meet error: %s", str(ex))

    def _generate(
            self,
            prompts: List[str],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> LLMResult:
        """Run the LLM on the given prompt and input."""
        generations = []
        llm_output = {}
        for prompt in prompts:
            request_body = {
                "prompt": prompt,
                "stream": False,
                **get_llm_params({
                    "temperature": self.temperature,
                    "max_tokens": self.max_tokens,
                    "top_p": self.top_p
                })
            }
            token = self.token_getter.get_valid_token()
            headers = {
                AUTH_TOKEN_HEADER: token,
                "X-Agent": "pangu-kits-app-dev"
            } if token else {"X-Agent": "pangu-kits-app-dev"}
            rsp = requests.post(self.gallery_url, headers=headers,
                                json=request_body,
                                verify=False, stream=False, proxies=self.proxies)

            if 200 == rsp.status_code:
                llm_output = rsp.json()
                text = llm_output["choices"][0]["text"]
                generations.append([Generation(text=text)])
            else:
                logger.error("Call gallery llm failed, http status: %d, error response: %s",
                             rsp.status_code, rsp.content)
                rsp.raise_for_status()

        return LLMResult(generations=generations, llm_output=llm_output)


class GalleryChatLLM(BaseChatModel):
    temperature: Optional[float]
    max_tokens: Optional[int]
    top_p: Optional[float]
    gallery_url: str
    token_getter: IAMTokenProvider
    streaming: Optional[bool]
    proxies: dict = {}

    @property
    def _llm_type(self) -> str:
        return "gallery_chat_llm"

    def _combine_llm_outputs(self, llm_outputs: List[Optional[dict]]) -> dict:
        return llm_outputs[0]

    def _stream(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        history = self._get_history(messages)
        request_body = {
            "prompt": messages[-1].content,
            "stream": True,
            **get_llm_params({
                "temperature": self.temperature,
                "max_tokens": self.max_tokens,
                "top_p": self.top_p
            }),
            "history": history
        }
        token = self.token_getter.get_valid_token()
        headers = {
            AUTH_TOKEN_HEADER: token,
            "X-Agent": "pangu-kits-app-dev"
        } if token else {"X-Agent": "pangu-kits-app-dev"}
        logger.info("req: %s", json.dumps(request_body, ensure_ascii=False))
        rsp = requests.post(self.gallery_url, headers=headers,
                            json=request_body,
                            verify=False, stream=True, proxies=self.proxies)
        try:
            rsp.raise_for_status()
            stream_client: sseclient.SSEClient = sseclient.SSEClient(rsp)
            for event in stream_client.events():
                # 解析出Token数据
                data_json = json.loads(event.data)
                if data_json.get("choices") is None:
                    raise ValueError(f"Meet json decode error: {str(data_json)}, not get choices")
                chunk = ChatGenerationChunk(message=
                                            AIMessageChunk(content=data_json["choices"][0]["text"]))
                yield chunk
                if run_manager:
                    run_manager.on_llm_new_token(chunk.text)
        except JSONDecodeError as ex:
            # [DONE]表示stream结束了
            if event.data != "[DONE]":
                logger.warning("Meet json decode error: %s", str(ex))
        except ChunkedEncodingError as ex:
            logger.warning("Meet error: %s", str(ex))

    def _generate(self, messages: List[BaseMessage], stop: Optional[List[str]] = None,
                  run_manager: Optional[CallbackManagerForLLMRun] = None, **kwargs: Any) -> ChatResult:
        """Run the LLM on the given prompt and input."""
        history = self._get_history(messages)
        generations = []
        llm_output = {}
        request_body = {
            "prompt": messages[-1].content,
            "stream": False,
            **get_llm_params({
                "temperature": self.temperature,
                "max_tokens": self.max_tokens,
                "top_p": self.top_p
            }),
            "history": history
        }
        token = self.token_getter.get_valid_token()
        headers = {
            AUTH_TOKEN_HEADER: token,
            "X-Agent": "pangu-kits-app-dev"
        } if token else {"X-Agent": "pangu-kits-app-dev"}
        logger.info("req: %s", json.dumps(request_body, ensure_ascii=False))
        rsp = requests.post(self.gallery_url, headers=headers,
                            json=request_body,
                            verify=False, stream=False, proxies=self.proxies)

        if 200 == rsp.status_code:
            llm_output = rsp.json()
            text = llm_output["choices"][0]["text"]
            generations.append(ChatGeneration(message=AIMessage(content=text)))
        else:
            logger.error("Call gallery llm failed, http status: %d, error response: %s", rsp.status_code, rsp.content)
            rsp.raise_for_status()

        return ChatResult(generations=generations, llm_output=llm_output)

    @staticmethod
    def _get_history(messages: List[BaseMessage]) -> List:
        # 第一个role开始作为问答对开始
        message_req = convert_message_to_req(messages)
        while message_req and message_req[0].get("role") != "user":
            messages.pop(0)

        if not message_req:
            raise ValueError("No input messages!")
        history = []
        for i in range(0, len(message_req) - 1, 2):
            if message_req[i].get("role") != "user" and message_req[i+1].get("role") != "assistant":
                raise ValueError("Illegal question and answer pair!")
            history.append([message_req[i].get("content"), message_req[i+1].get("content")])
        return history


class GalleryLLMApi(AbstractLLMApi):

    def do_create_chat_llm(self, llm_config: LLMConfig) -> BaseChatModel:
        config_params = self._parse_llm_config(llm_config)
        return GalleryChatLLM(**config_params)

    def do_create_llm(self, llm_config: LLMConfig) -> BaseLLM:
        config_params = self._parse_llm_config(llm_config)
        return GalleryLLM(**config_params)

    def parse_llm_response(self, llm_result: LLMResult) -> LLMRespGallery:
        answer = llm_result.generations[0][0].text
        llm_output = llm_result.llm_output
        choices = [GalleryTextChoice(index=choice["index"],
                                     text=choice["text"]) for choice in llm_output.get("choices")]
        if llm_output.get("usage") is not None:
            usage = GalleryUsage(completion_tokens=llm_output.get("usage").get("completion_tokens"),
                                 prompt_tokens=llm_output.get("usage").get("prompt_tokens"),
                                 total_tokens=llm_output.get("usage").get("total_tokens"))
        else:
            usage = None

        llm_resp = LLMRespGallery(answer=answer,
                                  is_from_cache=False,
                                  gallery_text_resp=GalleryTextResp(id=llm_output.get("id"),
                                                                    created=llm_output.get("created"),
                                                                    choices=choices,
                                                                    usage=usage))
        return llm_resp

    def _parse_llm_config(self, llm_config: LLMConfig):
        config_params = {
            "gallery_url": llm_config.gallery_config.gallery_url,
            "token_getter": IAMTokenProviderFactory.create(llm_config.gallery_config.iam_config),
            "proxies": llm_config.gallery_config.http_config.requests_proxies()
        }
        llm_param_config_dict: dict = llm_config.llm_param_config.dict(exclude_none=True, exclude={"stream"})
        config_params = {**config_params, **llm_param_config_dict}
        return config_params



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from langchain_pangu.pangukitsappdev.api.schema import LLMResp
from langchain_pangu.pangukitsappdev.llms.response.pangu_text_resp import PanguTextResp


class LLMRespPangu(LLMResp):
    """Pangu response封装结构体
    Attributes:
        pangu_text_resp: api返回response结构体
    """
    pangu_text_resp: PanguTextResp




#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.

from typing import Optional, Any, Dict, Union, List

from langchain.callbacks.manager import CallbackManagerForChainRun
from langchain.chains import RetrievalQAWithSourcesChain


class RetrievalQAWithAllSourcesChain(RetrievalQAWithSourcesChain):
    """
    不依赖LLM的输出获取答案的来源，而是根据查询出来的doc数据直接获取source信息
    """

    @property
    def _chain_type(self) -> str:
        return "RetrievalQAWithAllSourcesChain"

    def _call(
            self,
            inputs: Dict[str, Any],
            run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, Union[str, List]]:
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        docs = self._get_docs(inputs)
        answer = self.combine_documents_chain.run(
            input_documents=docs, callbacks=_run_manager.get_child(), **inputs
        )

        result: Dict[str, Any] = {
            self.answer_key: answer,
            self.sources_answer_key: [d.metadata for d in docs],
        }
        if self.return_source_documents:
            result["source_documents"] = docs
        return result



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from langchain_pangu.pangukitsappdev.api.schema import LLMResp
from langchain_pangu.pangukitsappdev.llms.response.openai_text_resp import OpenAITextResp


class LLMRespOpenAI(LLMResp):
    """OpenAI response封装结构体
    Attributes:
        openai_text_resp: openai sdk返回结构体
    """
    openai_text_resp: OpenAITextResp




#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import List

from langchain_pangu.pangukitsappdev.api.memory.vector.base import VectorApi
from langchain_pangu.pangukitsappdev.api.memory.vector.factory import Vectors
from langchain_pangu.pangukitsappdev.api.memory.vector.vector_config import VectorStoreConfig
from langchain_pangu.pangukitsappdev.api.retriever.base import AbstractToolRetriever
from langchain_pangu.pangukitsappdev.api.tool.base import AbstractTool
from langchain_pangu.pangukitsappdev.retriever.tool_metadata import ToolMetadata
from langchain_pangu.pangukitsappdev.tool.tool_provider import ToolProvider


class ToolRetrieverImpl(AbstractToolRetriever):

    def __init__(self, tool_provider: ToolProvider, vector_store_config: VectorStoreConfig):
        """储存工具类相关信息
        初始化
        :param tool_provider: 工具存储
        :param: vector_store_config: 语义向量存储配置
        """
        super().__init__(tool_provider, vector_store_config)
        self.vector: VectorApi = Vectors.of(vector_config=vector_store_config)

    def add_tools(self, tools: List[AbstractTool]):
        if not self.vector_store_config.vector_fields:
            raise ValueError("vector_store_config must specify vector_fields!")
        tool_metadata_list = [ToolMetadata(tool_id=tool.get_tool_id(),
                                           tool_metadata={
                                               self.vector_store_config.vector_fields[0]:
                                                   tool.get_tool_desc() + "，" + tool.principle})
                              for tool in tools]
        self.add_tools_from_metadata(tool_metadata_list)

    def dsl_search(self, query: str, dsl: str) -> List[AbstractTool]:
        raise NotImplementedError("Unsupported operation!")



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import logging
from abc import ABC, abstractmethod
from typing import List, Dict, Any

from langchain.schema.prompt_template import BasePromptTemplate

from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi
from langchain_pangu.pangukitsappdev.api.memory.vector.base import Document
from langchain_pangu.pangukitsappdev.api.skill.base import Skill

logger = logging.getLogger(__name__)

DOCUMENTS_VAR_KEY = "documents"
QUESTION_VAR_KEY = "question"


class AbstractDocSkill(Skill, ABC):
    """文档处理Skill
    """

    @abstractmethod
    def execute_with_docs(self, docs: List[Document], question: str = None) -> str:
        """文档摘要/问答
        Args:
            docs: 文档
            question: (Optional)问题，如果不传递question，则对文档进行摘要，否则就是文档文档

        Returns:
            回答或者摘要
        """
        pass

    def execute_with_texts(self, texts: List[str], question: str = None) -> str:
        """文档摘要/问答
        Args:
            texts: 文档内容
            question: (Optional)问题，如果不传递question，则对文档进行摘要，否则就是文档文档

        Returns:
            回答或者摘要
        """
        return self.execute_with_docs([Document(page_content=t) for t in texts], question=question)

    def execute(self, inputs: Dict[str, Any]) -> str:
        return self.execute_with_docs(docs=inputs.get(DOCUMENTS_VAR_KEY), question=inputs.get(QUESTION_VAR_KEY))


class DocStuffSkill(AbstractDocSkill):
    """
    stuff策略
    """

    def __init__(self, prompt_template: BasePromptTemplate, llm_api: LLMApi):
        self.prompt_template = prompt_template
        self.llm_api = llm_api

    def execute_with_docs(self, docs: List[Document], question: str = None) -> str:
        prompt = self.prompt_template.format(**{DOCUMENTS_VAR_KEY: [{"pageContent": d.page_content} for d in docs],
                                                QUESTION_VAR_KEY: question})
        return self.skill_llm_ask(prompt, self.llm_api)


class DocRefineSkill(AbstractDocSkill):
    """
    refine策略

    Attributes:
        qa_prompt: 问答的prompt模板
        refine_prompt: 精炼的prompt模板
        llm_api: llm
    """

    def __init__(self, qa_prompt: BasePromptTemplate, refine_prompt: BasePromptTemplate, llm_api: LLMApi):
        self.qa_prompt = qa_prompt
        self.refine_prompt = refine_prompt
        self.llm_api = llm_api

    def execute_with_docs(self, docs: List[Document], question: str = None) -> str:
        prompt = self.qa_prompt.format(document=docs[0].page_content, question=question)
        answer = self.skill_llm_ask(prompt, self.llm_api)

        logger.debug("Refine policy first answer %s", answer)

        if len(docs) == 1:
            return answer

        for doc in docs[1:]:
            refine_prompt = self.refine_prompt.format(document=doc.page_content,
                                                      question=question,
                                                      answer=answer)
            answer = self.skill_llm_ask(refine_prompt, self.llm_api)
            logger.debug("Refine policy refined answer: %s", answer)

        return answer


class DocMapReduceSkill(AbstractDocSkill):
    """MapReduce策略
    Attributes:
        map_prompt: Map阶段使用的prompt模板
        reduce_prompt: reduce阶段使用的prompt模板
        llm_api: 大语言模型的接口
        reduce_max_token: （Optional） 默认每次组合reduce最大的token数量，防止超过LLM的参数限制
    """

    def __init__(self, map_prompt: BasePromptTemplate,
                 reduce_prompt: BasePromptTemplate,
                 llm_api: LLMApi,
                 reduce_max_token: int = 2000):
        self.map_prompt = map_prompt
        self.reduce_prompt = reduce_prompt
        self.llm_api = llm_api
        self.reduce_max_token = reduce_max_token

    def execute_with_docs(self, docs: List[Document], question: str = None) -> str:
        def map_doc(doc: Document) -> Document:
            prompt = self.map_prompt.format(document=doc.page_content, question=question)
            answer = self.skill_llm_ask(prompt, self.llm_api)
            logger.debug("policy: mapreduce, document is: %s", doc.page_content)
            logger.debug("policy: mapreduce, map answer is: %s", answer)
            return Document(page_content=answer)

        def format_reduce(docs_group: List[Document]) -> str:
            return self.reduce_prompt.format(summaries=[{"pageContent": d.page_content} for d in docs_group],
                                             question=question)

        def reduce_docs(docs_group: List[Document]) -> str:
            return self.skill_llm_ask(format_reduce(docs_group), self.llm_api)

        mapped_docs = [map_doc(doc) for doc in docs]

        while len(format_reduce(mapped_docs)) > self.reduce_max_token:
            docs_groups = []
            sub_docs = []
            for doc in mapped_docs:
                sub_docs.append(doc)
                if len(format_reduce(sub_docs)) < self.reduce_max_token:
                    continue

                if len(sub_docs) <= 2:
                    raise ValueError("single document is too long")

                docs_groups.append(sub_docs[0:-1])
                sub_docs.clear()
                sub_docs.append(doc)

            if sub_docs:
                docs_groups.append(sub_docs)

            mapped_docs.clear()

            mapped_docs = [Document(page_content=reduce_docs(docs_group)) for docs_group in docs_groups]

        return reduce_docs(mapped_docs)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from __future__ import unicode_literals

import json
import logging
import uuid
from abc import ABC, abstractmethod
from json import JSONDecodeError
from typing import List, Union, Optional, Dict, Any

from langchain.schema.messages import AIMessageChunk, BaseMessage
from langchain.schema.output import ChatGenerationChunk, LLMResult
from pydantic.v1.json import pydantic_encoder

from langchain_pangu.pangukitsappdev.agent.agent_action import AgentAction
from langchain_pangu.pangukitsappdev.agent.agent_session import AgentSession
from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi, ConversationMessage, Role
from langchain_pangu.pangukitsappdev.api.llms.llm_config import LLMModuleConfig
from langchain_pangu.pangukitsappdev.api.retriever.base import ToolRetriever
from langchain_pangu.pangukitsappdev.api.tool.base import AbstractTool
from langchain_pangu.pangukitsappdev.callback.StreamCallbackHandler import StreamCallbackHandler

logger = logging.getLogger(__name__)


class AgentListener(ABC):
    """Agent监听，允许对Agent的各个阶段进行处理
    """

    def on_session_start(self, agent_session: AgentSession):
        """
        Session启动时调用
        :param agent_session: AgentSession
        """

    def on_session_iteration(self, agent_session: AgentSession):
        """
        Session迭代过程中调用
        :param agent_session: AgentSession
        """

    def on_session_end(self, agent_session: AgentSession):
        """
        Session结束时调用
        :param agent_session: AgentSession
        """

    def on_check_interrupt_requirements(self, agent_session: AgentSession):
        """
        onSessionIteration调用结束后，检查Agent是否需要终止，如果需要终止，则返回true，默认不终止
        可以在终止前对agentSession进行修改，如：修改agent的finalAnswer
        :param agent_session: AgentSession
        :return: bool类型结果
        """
        return False


class Agent(ABC):

    @abstractmethod
    def add_tool(self, tool: AbstractTool):
        """
        为Agent增加工具类
        :param tool: Tool
        """

    @abstractmethod
    def remove_tool(self, tool_id: str):
        """
        删除tool
        :param tool_id: 工具id
        """

    @abstractmethod
    def clear_tool(self):
        """
        删除所有工具
        """

    @abstractmethod
    def run_step(self, agent_session: AgentSession) -> AgentSession:
        """
        单步执行Agent，提供干预能力
        :param agent_session: 初始状态或运行中状态的agentSession，包括历史消息及其执行步骤，可以使用AgentSessionHelper类辅助处理
        :return: Agent执行的结果
        """

    @abstractmethod
    def run(self, user_input: Union[List[ConversationMessage], str, AgentSession]) -> AgentSession:
        """
        递归执行Agent，直到完成所有步骤的执行
        :param user_input: 用户的输入, 支持多轮List[ConversationMessage]或str或AgentSession
        :return: 计划的结果
        """

    @abstractmethod
    def set_max_iterations(self, iterations: int):
        """
        设置最大迭代次数
        :param iterations: 次数
        """

    @abstractmethod
    def set_tool_retriever(self, tool_retriever: ToolRetriever):
        """
        设置工具检索器
        :param tool_retriever: 工具检索器
        """

    @abstractmethod
    def add_listener(self, agent_listener: AgentListener):
        """
        添加一个Agent监听器
        :param agent_listener: Agent监听器
        """

    @abstractmethod
    def set_stream_callback(self, text_stream_callback: StreamCallbackHandler,
                            tool_stream_callback: StreamCallbackHandler):
        """
        设置流式输出回调函数
        :param text_stream_callback: Agent文本输出的StreamCallback
        :param tool_stream_callback: Agent工具输出的StreamCallBack
        """


class AbstractAgent(Agent):

    def __init__(self, llm: LLMApi):
        """
        构造一个agent
        :param llm: LLMApi
        """
        self.llm = llm
        self.tool_map: Dict[str, AbstractTool] = {}
        self.max_iterations = 15
        self.agent_listener: Optional[AgentListener] = None
        self.tool_retriever: Optional[ToolRetriever] = None

    @staticmethod
    def convert_message_to_dict(conversation_messages: List[ConversationMessage]) -> List[Dict]:
        # 用于将python命名风格变量适配小驼峰，与Java版本保持prompt模板统一
        return [
            {
                "role": {
                    "text": message.role.text,
                    "desc": message.role.desc
                },
                "actions": [
                    {
                        "req": action.req,
                        "resp": action.resp,
                        "thought": action.thought,
                        "actionJson": action.action_json,
                        "action": action.action,
                        "actionInput": action.action_input,
                        "observation": action.observation,
                        "userFeedBack": action.user_feedback
                    } for action in message.actions
                ],
                "content": message.content
            } for message in conversation_messages]

    def add_tool(self, tool: AbstractTool):
        tool_id = tool.get_tool_id()
        self.tool_map.update({tool_id: tool})

    def remove_tool(self, tool_id: str):
        self.tool_map.pop(tool_id)

    def clear_tool(self):
        self.tool_map.clear()

    def run_step(self, agent_session: AgentSession) -> AgentSession:
        # tool检索器处理
        self.init_tool_from_retriever(agent_session.messages)
        # 关闭自动添加新的SystemMessage
        self.llm.get_llm_config().llm_module_config.enable_append_system_message = False
        agent_session.is_by_step = True
        # react 单步执行
        try:
            self.react(agent_session)
        except (ValueError, JSONDecodeError, TypeError) as e:
            logger.debug("run error when call react", e)
            raise e

        if agent_session.agent_session_status != "FINISHED":
            self.notice_session_iteration(agent_session, agent_session.current_action)
        return agent_session

    def run(self, user_input: Union[List[ConversationMessage], str, AgentSession]) -> AgentSession:
        """
        执行agent
        :param user_input: 用户的输入, 支持多轮List[ConversationMessage]或str或AgentSession
        :return: 计划的结果
        """
        agent_session = AgentSessionHelper.init_agent_session(user_input) if \
            isinstance(user_input, str) or isinstance(user_input, List) else user_input
        # tool检索器处理
        self.init_tool_from_retriever(agent_session.messages)
        # 关闭自动添加新的SystemMessage
        self.llm.get_llm_config().llm_module_config.enable_append_system_message = False
        self.notice_session_start(agent_session)
        agent_session.is_by_step = False
        # 新增一个Assistant回复
        if agent_session.messages[-1].role.text != "assistant":
            agent_session.current_message = ConversationMessage(role=Role.ASSISTANT, content="")
            agent_session.messages.append(agent_session.current_message)
        try:
            self.react(agent_session)
        except (ValueError, JSONDecodeError, TypeError) as e:
            logger.debug("run error when call react", e)
            raise e

        # 设置Assistant消息内容
        AgentSessionHelper.update_assistant_message(agent_session, False)
        actions = agent_session.current_message.actions
        if actions:
            logger.info(AgentSessionHelper.print_plan(agent_session))
        return agent_session

    @abstractmethod
    def react(self, agent_session: AgentSession):
        """
        迭代解决问题
        :param agent_session: 历史迭代几率
        """

    def add_listener(self, agent_listener: AgentListener):
        self.agent_listener = agent_listener

    def set_max_iterations(self, iterations: int):
        if iterations <= 0:
            raise ValueError("iterations value not legal.")
        self.max_iterations = iterations

    def set_tool_retriever(self, tool_retriever: ToolRetriever):
        self.tool_retriever = tool_retriever

    def notice_session_start(self, agent_session: AgentSession):
        agent_session.agent_session_status = "RUNNING"
        if self.agent_listener:
            self.agent_listener.on_session_start(agent_session)

    def notice_session_iteration(self, agent_session: AgentSession, action: AgentAction):
        agent_session.current_message.actions.append(action)
        agent_session.agent_session_status = "RUNNING"

        if self.agent_listener:
            self.agent_listener.on_session_iteration(agent_session)

    def notice_session_end(self, agent_session: AgentSession, action: AgentAction):
        agent_session.current_message.actions.append(action)
        agent_session.agent_session_status = "FINISHED"
        if self.agent_listener:
            self.agent_listener.on_session_end(agent_session)

    def tool_execute(self, tool: AbstractTool, tool_input: Union[str, dict], agent_session: AgentSession):
        if agent_session.is_by_step:
            return
        try:
            tool_result = tool.run(tool_input)
        except TypeError:
            tool_result = tool.run(str(tool_input))
        action = agent_session.current_action
        if isinstance(tool_result, (str, int, float, bool)):
            action.observation = str(tool_result)
        else:
            action.observation = json.dumps(tool_result, default=pydantic_encoder, ensure_ascii=False)
        # 本次迭代结束
        self.notice_session_iteration(agent_session, action)
        # 执行下一迭代
        self.react(agent_session)

    @staticmethod
    def sub_str_between(origin_str: str, start_str: str, end_str: str):
        if origin_str:
            start_pos = origin_str.find(start_str)
            if start_pos != -1:
                end_pos = origin_str.find(end_str)
                if end_pos != -1:
                    return origin_str[start_pos + len(start_str): end_pos]
        return ""

    @staticmethod
    def sub_str_before(origin_str: str, separator: str):
        if origin_str:
            if not separator:
                return ""
            else:
                pos = origin_str.find(separator)
                return origin_str if pos == -1 else origin_str[:pos]
        else:
            return origin_str

    @staticmethod
    def sub_str_after(origin_str: str, separator: str):
        if origin_str:
            if separator == "":
                return ""
            else:
                pos = origin_str.find(separator)
                return "" if pos == -1 else origin_str[pos + len(separator):]
        else:
            return origin_str

    @staticmethod
    def remove_start(origin_str: str, remove: str):
        if origin_str and remove:
            return origin_str[len(remove):] if origin_str.startswith(remove) else origin_str
        else:
            return origin_str

    def need_interrupt(self, agent_session: AgentSession) -> bool:
        if not agent_session.current_message.actions:
            return False
        # 超过最大迭代次数终止
        if len(agent_session.current_message.actions) >= self.max_iterations:
            logger.debug("stopped due to iteration limit. maxIterations is %s", self.max_iterations)
            return True
        # 用户终止
        if self.agent_listener and self.agent_listener.on_check_interrupt_requirements(agent_session):
            agent_session.agent_session_status = "INTERRUPTED"
            logger.info("agent stopped due to manual interruption")
            return True
        return False

    def init_tool_from_retriever(self, messages: List[ConversationMessage]):
        if self.tool_retriever is None or messages is None:
            return
        # 处理多轮对话
        preprocessor = self.tool_retriever.get_query_preprocessor()
        query = preprocessor(messages)
        tools = self.tool_retriever.search(query)
        self.tool_map.clear()
        for tool in tools:
            self.add_tool(tool)

    def get_system_prompt(self, agent_session: AgentSession):
        # 优先取设置到的LLM人设
        system_prompt = self.llm.get_llm_config().llm_module_config.system_prompt
        if system_prompt is not None:
            return system_prompt
        # 其次取message中最后一个人设
        for message in reversed(agent_session.messages):
            if message.role.text == "system":
                return message.content
        return None

    def set_stream_callback(self, text_stream_callback: StreamCallbackHandler,
                            tool_stream_callback: StreamCallbackHandler):
        self.llm.set_callback(AgentStreamCallBack(text_stream_callback=text_stream_callback,
                                                  tool_stream_callback=tool_stream_callback,
                                                  llm_module_config=self.llm.get_llm_config().llm_module_config))

    def get_tool(self, tool_id: str):
        if not self.tool_map:
            logger.error("there is no tool in agent")
            raise ValueError("there is no tool in agent")
        tool = self.tool_map.get(tool_id)
        if not tool:
            # 如果没有合适tool，则使用最相似的tool
            tool_map_keys = list(self.tool_map.keys())
            distance_list = list(map(AbstractAgent.levenshtein_distance,
                                     [tool_id for _ in range(len(tool_map_keys))],
                                     tool_map_keys))
            new_tool_id = tool_map_keys[distance_list.index(min(distance_list))]
            logger.warning("can not find tool for %s in %s, the most similar tool is %s",
                           tool_id, tool_map_keys, new_tool_id)
            return self.tool_map.get(new_tool_id)
        return tool

    @staticmethod
    def levenshtein_distance(tool_id: str, tool_map_key: str) -> int:
        if not tool_id or not tool_map_key:
            raise ValueError("tool_id must not be empty or None")
        if len(tool_id) < len(tool_map_key):
            left = tool_id
            right = tool_map_key
        else:
            left = tool_map_key
            right = tool_id
        left_len = len(left)
        right_len = len(right)
        # dp数组初始化
        dp = [[right_len for _ in range(right_len + 1)] for _ in range(left_len + 1)]
        for i in range(left_len + 1):
            dp[i][0] = i
        for i in range(right_len + 1):
            dp[0][i] = i
        for i in range(1, left_len + 1):
            for j in range(1, right_len + 1):
                cost = 0 if left[i - 1] == right[j - 1] else 1
                dp[i][j] = min(min(dp[i - 1][j] + 1, dp[i][j - 1] + 1), dp[i - 1][j - 1] + cost)
        return dp[left_len][right_len]


class AgentStreamCallBack(StreamCallbackHandler):
    def __init__(self, text_stream_callback: StreamCallbackHandler,
                 tool_stream_callback: StreamCallbackHandler,
                 llm_module_config: LLMModuleConfig):
        super().__init__()
        self.text_stream_callback = text_stream_callback
        self.tool_stream_callback = tool_stream_callback
        self.llm_module_config = llm_module_config
        self.serialized = {}
        self.messages = []
        self.in_bool_stream = False
        self.tool_generation: Optional[ChatGenerationChunk] = None

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        **kwargs: Any,
    ) -> Any:
        self.text_stream_callback.on_chat_model_start(serialized=serialized, messages=messages, **kwargs)
        self.serialized = serialized
        self.messages = messages

    def on_llm_error(
        self,
        error: BaseException,
        **kwargs: Any,
    ) -> Any:
        self.text_stream_callback.on_llm_error(error=error, **kwargs)
        self.tool_stream_callback.on_llm_error(error=error, **kwargs)

    def on_llm_end(
        self,
        response: LLMResult,
        **kwargs: Any,
    ) -> Any:
        self.text_stream_callback.on_llm_end(response=response, **kwargs)

    def on_llm_new_token(
        self,
        token: str,
        **kwargs: Any,
    ) -> Any:
        unify_tool_tag_prefix = self.llm_module_config.llm_module_property.unify_tool_tag_prefix
        unify_tool_tag_suffix = self.llm_module_config.llm_module_property.unify_tool_tag_suffix
        if unify_tool_tag_prefix in token:
            # 工具调用之前部分使用文本流输出
            text_token = AbstractAgent.sub_str_before(token, unify_tool_tag_prefix)
            self.text_stream_callback.on_llm_new_token(token=text_token, **kwargs)
            # 工具调用开始，工具调用后部分使用工具流输出
            self.tool_stream_callback.on_chat_model_start(serialized=self.serialized, messages=self.messages, **kwargs)
            self.in_bool_stream = True
            tool_token = AbstractAgent.sub_str_after(token, unify_tool_tag_prefix)
            self.tool_stream_callback.on_llm_new_token(token=tool_token, **kwargs)
            self.tool_generation = ChatGenerationChunk(message=AIMessageChunk(content=tool_token))
        elif unify_tool_tag_suffix in token:
            # 工具调用结束之前部分仍用工具流输出
            tool_token = AbstractAgent.sub_str_before(token, unify_tool_tag_suffix)
            self.tool_stream_callback.on_llm_new_token(token=tool_token, **kwargs)
            self.tool_generation += ChatGenerationChunk(message=AIMessageChunk(content=tool_token))
            # 工具调用结束
            self.tool_stream_callback.on_llm_end(response=LLMResult(generations=[[self.tool_generation]]))
            self.in_bool_stream = False
            # 工具调用后部分文本流输出
            text_token = AbstractAgent.sub_str_after(token, unify_tool_tag_suffix)
            self.text_stream_callback.on_llm_new_token(token=text_token, **kwargs)
        else:
            if self.in_bool_stream:
                # 工具流
                self.tool_stream_callback.on_llm_new_token(token=token, **kwargs)
                self.tool_generation += ChatGenerationChunk(message=AIMessageChunk(content=token))
            else:
                # 文本流
                self.text_stream_callback.on_llm_new_token(token=token, **kwargs)


class AgentSessionHelper:
    """
    AgentSession辅助类
    """
    FINAL_ACTION = "FINAL_ANSWER"

    @staticmethod
    def init_agent_session(user_message: Union[str, List[ConversationMessage]]) -> AgentSession:
        """
        使用用户消息初始化一个AgentSession
        :param user_message: 用户消息
        :return: agent_session
        """
        messages = [ConversationMessage(role=Role.USER,
                                        content=user_message)] if isinstance(user_message, str) else user_message
        agent_session = AgentSession(messages=messages, session_id=str(uuid.uuid4()),
                                     history_action=[], agent_session_status="INIT")
        # 新增一个Assistant回复
        agent_session.current_message = ConversationMessage(role=Role.ASSISTANT, content="")
        agent_session.messages.append(agent_session.current_message)
        return agent_session

    @staticmethod
    def update_assistant_message(agent_session: AgentSession, override: bool):
        """
        使用当前的action更新AssistantMessage的内容
        :param agent_session: AgentSession
        :param override: 是否覆盖已有的内容
        """
        if override or not agent_session.current_message.content:
            agent_session.current_message.content = str(agent_session.current_action.action_input)


    @staticmethod
    def set_tool_output(agent_session: AgentSession, observation: str):
        """
        向Agent的当前步骤设置工具返回结果
        :param agent_session: AgentSession
        :param observation: 工具返回结果
        """
        agent_session.current_action.observation = observation

    @staticmethod
    def set_user_feedback(agent_session: AgentSession, user_feedback: str):
        """
        向Agent的当前步骤设置用户反馈
        :param agent_session: AgentSession
        :param user_feedback: 用户反馈
        """
        agent_session.current_action.user_feedback = user_feedback

    @staticmethod
    def print_plan(agent_session: AgentSession) -> str:
        """
        打印AgentSession
        :param agent_session: AgentSession
        :return: AgentSession的字符串输出
        """
        log_msg = ""
        for message in agent_session.messages:
            log_msg += f"\n{message.role.desc}: {message.content}"
            # 如果为Assistant消息，打印可能存在的Action
            if message.role.text == "assistant" and message.actions:
                actions = message.actions
                for i in range(len(actions)):
                    log_msg += f"\n - 步骤{i+1}"
                    action = actions[i]
                    if action.thought and action.thought != action.action_input:
                        thought = action.thought.replace("\n", "")
                        log_msg += f":\n   思考:{thought}"
                    if action.action == "FINAL_ANSWER":
                        log_msg += f"\n   答复:{action.action_input}"
                    else:
                        log_msg += f"\n   行动:使用工具[{action.action}],传入参数{action.action_input}"
                        if action.observation:
                            log_msg += f"\n   工具返回:{action.observation}"
                        if action.user_feedback:
                            log_msg += f"\n   用户反馈:{action.user_feedback}"
        return log_msg



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import base64
import hmac
import secrets
from hashlib import sha256


class SecurityUtil:

    @staticmethod
    def get_security_random(bits: int = 256) -> str:
        """
        返回当前的时间戳毫秒值
        :param bits: 随机数的bit位数（取8的倍数，如256）
        :return: 十六进制字符串
        """
        try:
            return str(secrets.token_bytes(int(bits / 8)).hex())
        except ValueError:
            raise ValueError("get secure random exception.")

    @staticmethod
    def hmac_sha256_base64(msg: str, key_str: str) -> str:
        """
        进行hmacSHA256的加密，返回base64的字符串
        :param msg: 加密明文
        :param key_str: 密钥,经过HEX编码后的String串
        :return: Hex后的字符串
        """
        hmac_msg = msg.encode('utf-8')
        hmac_key = key_str.encode('utf-8')
        signature = base64.b64encode(hmac.new(hmac_key, hmac_msg, digestmod=sha256).digest()).decode('utf-8')
        return signature



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import json
import logging
import os

from langchain.prompts import PromptTemplate

from langchain_pangu.pangukitsappdev.api.prompt.prompt_config import PromptConfig, PromptTemplatesFileConfig

logger = logging.getLogger(__name__)


def read_prompt_from(tmpl_file) -> PromptTemplate:
    with open(tmpl_file, mode="r", encoding="utf8") as ff:
        txt = ff.read()
    return PromptTemplate.from_template(txt, template_format="jinja2")


class PromptTemplates:
    """获取Prompt模板工厂类
        提供get方法根据名字获取prompt模板。提供register方法从文件中注册一个模板
    Attributes:
        prompt_tmpl_lookup: 类成员变量，维护prompt模板
    """

    prompt_tmpl_lookup: dict = {}

    @classmethod
    def get(cls, tmpl_name) -> PromptTemplate:
        """
        根据tmpl_name获取一个模板，优先自定义模板
        Args:
            tmpl_name: 唯一标识一个prompt模板

        Returns:
            prompt模板PromptTemplate

        """
        custom_config = PromptConfig()
        custom_prompt_path = custom_config.custom_prompt_path
        if prompt_files.get(tmpl_name) is None:
            return cls.prompt_tmpl_lookup.get(tmpl_name)
        if custom_prompt_path and os.path.exists(os.path.join(custom_prompt_path, prompt_files.get(tmpl_name))):
            return read_prompt_from(os.path.join(custom_prompt_path, prompt_files.get(tmpl_name)))
        else:
            return cls.prompt_tmpl_lookup.get(tmpl_name)

    @classmethod
    def register(cls, tmpl_file, tmpl_name=None):
        """从文件内容注册一个prompt模板
        文件的后缀代表这个模板的格式化类型，支持f-string和jinja2。比如有一个模板文件是foo.f-string，则会使用f"xx{}xxx"的方式做格式化
        Args:
            tmpl_file: 指向模板文件的路径。文件的后缀代表模板的格式化类型，支持f-string和jinja2
            tmpl_name: （Optional） 唯一指定一个模板的名称，如果不传则使用不包含后缀的文件名
        """
        local_tmpl_name = tmpl_name if tmpl_name else os.path.basename(tmpl_file).split(".")[0]

        cls.prompt_tmpl_lookup[local_tmpl_name] = read_prompt_from(tmpl_file)

    @classmethod
    def template_names(cls) -> list:
        """
        获取所有可以通过PromptTemplateProvider获取的prompt模板名称
        Returns: list, 所有模板名称
        """
        return list(cls.prompt_tmpl_lookup.keys())


# load prompt settings
prompt_config = PromptConfig()
prompt_file_config = PromptTemplatesFileConfig()

# load default prompt path
py_file_path = os.path.split(__file__)[0]
default_prompt_path = os.path.join(py_file_path, prompt_config.default_prompt_path)

# load prompt name for path dict by BaseSettings.json method
prompt_files = json.loads(prompt_file_config.json())

for key in prompt_files.keys():
    # register related prompts
    prompt_path = os.path.join(default_prompt_path, prompt_files[key])
    PromptTemplates.register(prompt_path, key)
logger.debug("Total load default prompts num: %s", len(prompt_files.keys()))



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from langchain.memory.chat_message_histories.redis import RedisChatMessageHistory

from langchain_pangu.pangukitsappdev.api.memory.cache.cache_config import ServerInfoRedis
from langchain_pangu.pangukitsappdev.api.memory.message_history_config import MessageHistoryConfig


class RedisMessageHistory(RedisChatMessageHistory):
    """继承 langchain RedisChatMessageHistory，根据 MessageHistoryConfig传参"""

    def __init__(self, msg_history_config: MessageHistoryConfig = None):
        if not msg_history_config:
            msg_history_config = MessageHistoryConfig()
        if not isinstance(msg_history_config.server_info, ServerInfoRedis):
            raise ValueError("Type of server_info should be ServerInfoRedis")
        super().__init__(
            session_id=msg_history_config.session_tag,
            url=msg_history_config.server_info.get_urls()[0],
            key_prefix=msg_history_config.key_prefix,
            ttl=msg_history_config.ttl
        )



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Any, Optional

from cachetools import TTLCache, LRUCache
from langchain.schema.cache import BaseCache
from langchain.schema.cache import RETURN_VAL_TYPE

from langchain_pangu.pangukitsappdev.api.memory.cache.base import CacheApiAdapter
from langchain_pangu.pangukitsappdev.api.memory.cache.cache_config import CacheStoreConfig


class InMemoryCacheApi(CacheApiAdapter):

    def __init__(self, cache_config: CacheStoreConfig):
        if cache_config.expire_after_write > 0:
            raise ValueError("InMemory Cache do not support expire_after_write")
        in_memory_cache = TTLInMemoryCache(cache_config.expire_after_access,
                                           cache_config.maximum_size)
        super().__init__(in_memory_cache, cache_config.session_tag)


class TTLInMemoryCache(BaseCache):
    def __init__(self, ttl, maximum_size):
        self.ttl = ttl
        max_size = maximum_size if maximum_size > 0 else float('inf')
        # 非Time-Based过期采用LRUCache
        if self.ttl <= 0:
            self.cache = LRUCache(maxsize=max_size)
        else:
            self.cache = TTLCache(maxsize=max_size, ttl=self.ttl)

    def lookup(self, prompt: str, llm_string: str) -> Optional[RETURN_VAL_TYPE]:
        return self.cache.get((prompt, llm_string))

    def update(self, prompt: str, llm_string: str, return_val: RETURN_VAL_TYPE) -> None:
        self.cache[(prompt, llm_string)] = return_val

    def clear(self, **kwargs: Any) -> None:
        self.cache.clear()



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Optional
from urllib.parse import quote

from pydantic.v1 import Field, validator

from langchain_pangu.pangukitsappdev.api.config_loader import SdkBaseSettings

AUTH_TOKEN_HEADER = "X-Auth-Token"
IAM_TOKEN_HEADER = "X-Subject-Token"


class HttpConfig(SdkBaseSettings):
    """http相关配置
    Attributes:
        proxy_default_enabled：默认代理开关
        proxy_enabled: 优先模块代理开关，模块未定义则使用默认代理开关
        proxy_url: 代理host地址:代理端口
        proxy_user: 认证用户
        proxy_passwd: 认证密码
    """

    def __init__(self, env_prefix="sdk.proxy", **kwargs):
        super().__init__(env_prefix=env_prefix, env_prefix_ident_fields=["enabled"], **kwargs)

    proxy_default_enabled: Optional[bool] = Field(env="sdk.proxy.enabled", default=False)
    proxy_enabled: Optional[bool] = Field(env="enabled")
    proxy_url: Optional[str] = Field(env="sdk.proxy.url")
    proxy_user: Optional[str] = Field(env="sdk.proxy.user")
    proxy_passwd: Optional[str] = Field(env="sdk.proxy.password")

    @validator("proxy_enabled")
    def get_proxy_enabled(cls, v, values):
        if v is None:
            v = values["proxy_default_enabled"]
        return v

    def get_proxy_url(self) -> str:
        if self.proxy_user and self.proxy_passwd:
            return f"{self.proxy_url.split('://')[0]}://{quote(self.proxy_user)}:{quote(self.proxy_passwd)}" \
                   f"@{self.proxy_url.split('://')[1]}"
        else:
            return f"{self.proxy_url}"

    def requests_proxies(self) -> dict:
        """
        构造requests请求时用的代理配置
        Returns:
            dict
        """
        return {
            "http": self.get_proxy_url(),
            "https": self.get_proxy_url(),
        } if self.proxy_enabled else {
            "http": "",
            "https": "",
        }


class IAMConfig(SdkBaseSettings):
    """
    IAM的配置参数
    """
    iam_disabled: Optional[bool] = Field(env="disabled")
    iam_ak: Optional[str] = Field(env="ak")
    iam_sk: Optional[str] = Field(env="sk")
    iam_url: Optional[str] = Field(env="url")
    iam_domain: Optional[str] = Field(env="domain")
    iam_user: Optional[str] = Field(env="user")
    iam_pwd: Optional[str] = Field(env="password")
    project_name: Optional[str] = Field(env="project")
    # token过期时间（毫秒），默认23小时，过期后，下一次接口调用会自动获取token
    expire_duration_millis: int = Field(default=23 * 60 * 60 * 1000, env="expireDurationMillis")
    x_auth_token: Optional[str] = Field(env="x_auth_token")
    http_config: HttpConfig = Field(default_factory=lambda: HttpConfig())


class IAMConfigWrapper:
    def __init__(self, env_prefix: str = "sdk.iam"):
        self.env_prefix = env_prefix

    def get_iam_config(self):
        sdk_iam = IAMConfig(env_prefix="sdk.iam",
                            http_config=HttpConfig(env_prefix=self.env_prefix.replace("iam", "proxy")))
        iam_config = IAMConfig(env_prefix=self.env_prefix,
                               http_config=HttpConfig(env_prefix=self.env_prefix.replace("iam", "proxy")))
        # 子模块设置iam_disabled使用，否则与全局iam_disabled一致
        if iam_config.iam_disabled is not None:
            sdk_iam.iam_disabled = iam_config.iam_disabled
        else:
            iam_config.iam_disabled = sdk_iam.iam_disabled
        if iam_config.iam_url is None and iam_config.x_auth_token is None:
            return sdk_iam
        else:
            return iam_config


class ServerInfo(SdkBaseSettings):
    """存储服务器链接相关信息
    Attributes:
        url: 主机ip列表:访问端口
        user: 用户
        password: 密钥
    """
    url: Optional[str] = Field(env="url")
    user: Optional[str] = Field(env="user")
    password: Optional[str] = Field(env="password")

    def has_auth(self) -> bool:
        return True if self.user and self.password else False

    def get_urls(self) -> [str]:
        raise ImportError("Not implemented")

    def get_http_urls(self) -> [str]:
        url_arr = self.url.split(",")
        return [
            f"{u.split('://')[0]}://{quote(self.user)}:{quote(self.password)}@{u.split('://')[1]}"
            for u in url_arr
        ] if self.has_auth() else [
            f"{u}" if 'http' in u else f"http://{u}"
            for u in url_arr
        ]


class OpenAIConfig(SdkBaseSettings):
    openai_base_url: Optional[str] = Field(env="sdk.llm.openai.url")
    openai_key: Optional[str] = Field(env="sdk.llm.openai.key")
    http_config: HttpConfig = Field(default_factory=lambda: HttpConfig(env_prefix="sdk.llm.openai.proxy"))



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Any, Optional

from pydantic.v1 import BaseModel


class AgentAction(BaseModel):
    """
    Agent执行的action
    Attributes:
        req: LLM的原始请求
        resp: LLM的原始响应
        thought: LLM的思考
        action_json: LLM输出的原始Action字符串
        action: LLM将要采取的行动，即Tool
        action_input: 行动的输入，即Tool的输入
        observation: 采取行动后给LLM观察的结果，即Tool的执行结果
        user_feedback: 用户对该步骤的反馈，可能需要更改工具调用参数
    """
    req: Optional[str]
    resp: Optional[str]
    thought: Optional[str]
    action_json: Optional[str]
    action: Optional[str]
    action_input: Optional[Any]
    observation: Optional[str]
    user_feedback: Optional[str]



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from typing import Dict

from langchain.prompts import PromptTemplate

from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi
from langchain_pangu.pangukitsappdev.api.llms.factory import LLMs
from langchain_pangu.pangukitsappdev.api.skill.base import SimpleSkill
from langchain_pangu.pangukitsappdev.skill.qa_with_sources import DocAskSkill


class DocumentSkills:
    """
    构造文档相关的Executor
    """

    prompt_template_map: Dict[str, PromptTemplate] = {}

    @classmethod
    def register_prompt_template(cls, skill_type, prompt_template: PromptTemplate):
        cls.prompt_template_map[skill_type] = prompt_template

    @classmethod
    def of(cls, llm_api: LLMApi) -> DocAskSkill:
        return DocAskSkill(llm_api)

    @classmethod
    def of_type(cls, skill_type: str, llm_api: LLMApi) -> DocAskSkill:
        pt = cls.prompt_template_map.get(skill_type)
        if not pt:
            raise ValueError(
                f"Unregistered skill type: {skill_type}, "
                f"please call register_prompt_template(skill_type, prompt_template) before use.")

        local_llm_api = llm_api if llm_api else LLMs.of("pangu")

        return SimpleSkill(local_llm_api, pt)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import logging

import urllib3

logging.captureWarnings(True)

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import logging
from typing import Dict, Any, Optional, List
from uuid import UUID

from langchain.callbacks.tracers import ConsoleCallbackHandler
from langchain.schema import BaseMessage


class DebugConsoleCallbackHandler(ConsoleCallbackHandler):
    """
    Only for debug，打印解析后的prompt数据供调试使用
    """

    def on_chat_model_start(self, serialized: Dict[str, Any], messages: List[List[BaseMessage]], *, run_id: UUID,
                            parent_run_id: Optional[UUID] = None, **kwargs: Any) -> Any:
        pass

    def on_text(self, text: str, *, run_id: UUID, parent_run_id: Optional[UUID] = None, **kwargs: Any) -> Any:
        super().on_text(text, run_id=run_id, parent_run_id=parent_run_id, **kwargs)
        logging.INFO(text)



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
from __future__ import unicode_literals

import json
import logging
from json import JSONDecodeError

from langchain.prompts import PromptTemplate

from langchain_pangu.pangukitsappdev.agent.agent_action import AgentAction
from langchain_pangu.pangukitsappdev.agent.agent_session import AgentSession
from langchain_pangu.pangukitsappdev.api.agent.base import AbstractAgent, AgentSessionHelper
from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi
from langchain_pangu.pangukitsappdev.prompt.prompt_tmpl import PromptTemplates

logger = logging.getLogger(__name__)


class ReactPanguAgent(AbstractAgent):
    DEFAULT_SYS_PROMPT = "你的名字叫智子，是由华为开发的智能助手"
    VERTICAL_SEPERATOR = "|"
    MODULE_VERSION_PREFIX_N2 = "N2"
    TEMPLATE_VERSION_AGENT_V2 = "agent_v2"
    TEMPLATE_VERSION_UNIFY = "unify"

    def __init__(self, llm: LLMApi):
        super(ReactPanguAgent, self).__init__(llm)
        self.set_default_unify_tag()

    def set_default_unify_tag(self):
        """
        给模型提供默认占位符
        """
        llm_module_property = self.llm.get_llm_config().llm_module_config.llm_module_property

        # 优先使用外部主动配置的参数
        if llm_module_property.unify_tool_tag_prefix and llm_module_property.unify_tool_tag_suffix and\
                llm_module_property.unify_tag_prefix and llm_module_property.unify_tag_suffix:
            return

        # 需要整体配置或不配置，不支持仅配置部分参数
        if llm_module_property.unify_tool_tag_prefix or llm_module_property.unify_tool_tag_suffix or\
                llm_module_property.unify_tag_prefix or llm_module_property.unify_tag_suffix:
            raise ValueError("Some unify tags are not configured.")

        # N2占位符为尖括号形式，后续版本为中括号形式
        if self.is_n2_module():
            llm_module_property.unify_tag_prefix = "<unused0>"
            llm_module_property.unify_tag_suffix = "<unused1>"
            llm_module_property.unify_tool_tag_prefix = "<unused2>"
            llm_module_property.unify_tool_tag_suffix = "<unused3>"
        else:
            llm_module_property.unify_tag_prefix = "[unused9]"
            llm_module_property.unify_tag_suffix = "[unused10]"
            llm_module_property.unify_tool_tag_prefix = "[unused11]"
            llm_module_property.unify_tool_tag_suffix = "[unused12]"

    def react(self, agent_session: AgentSession):
        # 超过最大迭代次数限制，不再执行
        if self.need_interrupt(agent_session):
            return

        # 构造React prompt
        react_tp = self.get_react_template()
        messages = agent_session.messages
        default_sys_prompt = self.DEFAULT_SYS_PROMPT if self.is_plugin_v1_version(self.get_template_version()) else ""
        sys_prompt = self.get_system_prompt(
            agent_session) if self.get_system_prompt(agent_session) is not None else default_sys_prompt
        final_prompt = react_tp.format(sys_prompt=sys_prompt,
                                       tool_desc=self.get_tool_desc(),
                                       messages=self.convert_message_to_dict(messages),
                                       cot_desc=self.llm.get_llm_config().llm_module_config.cot_desc)
        normalize_prompt = self.normalize_prompt_template(final_prompt)
        # 调用llm
        if self.llm.get_llm_config().llm_param_config.stream:
            tokens = self.llm.ask(normalize_prompt)
            answer = ""
            for token in tokens:
                answer += token
        else:
            answer = self.llm.ask(normalize_prompt).answer
        llm_module_property = self.llm.get_llm_config().llm_module_config.llm_module_property

        # 获取工具，例如：reserve_meeting_room|{'meetingRoom':'2303','start':'03:00','end':'08:00'}\n\n
        tool_use = self.sub_str_before(self.sub_str_between(answer,
                                                            llm_module_property.unify_tool_tag_prefix + "工具调用:",
                                                            llm_module_property.unify_tool_tag_suffix),
                                       llm_module_property.unify_tool_tag_prefix)
        # 新版本直接使用tool_tag_prefix，没有 工具调用: 的描述
        if not tool_use:
            tool_use = self.sub_str_before(self.sub_str_between(answer,
                                                                llm_module_property.unify_tool_tag_prefix,
                                                                llm_module_property.unify_tool_tag_suffix),
                                           llm_module_property.unify_tool_tag_prefix)
        tool_id = self.sub_str_before(tool_use, "|")
        # 未找到工具则返回
        if tool_id == "":
            action = AgentAction(req=normalize_prompt,
                                 resp=answer,
                                 thought=answer,
                                 action=AgentSessionHelper.FINAL_ACTION,
                                 action_input=answer)
            agent_session.current_action = action
            self.notice_session_end(agent_session, action)
            return
        tool = self.get_tool(tool_id)
        action = AgentAction(req=normalize_prompt,
                             resp=answer,
                             thought=self.sub_str_before(answer, llm_module_property.unify_tool_tag_prefix),
                             action_json="",
                             action=tool_id)
        agent_session.current_action = action

        # 提取工具参数
        action.action_input = self.sub_str_after(tool_use, self.VERTICAL_SEPERATOR).strip(self.VERTICAL_SEPERATOR)
        try:
            if tool.input_type in [int, float, str, bool]:
                json_obj = json.loads(action.action_input)
                if not json_obj or len(json_obj.values()) != 1:
                    raise ValueError(f"the action input is not a single input, require: {tool.get_pangu_function()},"
                                     f" action return: {action.action_input}")
                # 这里添加容错，对单个参数的字段名不做限制{}
                tool_input = list(json_obj.values())[0]
            elif tool.input_type is None:
                tool_input = "{}"
            else:
                tool_input = json.loads(action.action_input)
        except JSONDecodeError:
            tool_input = action.action_input

        # 执行工具
        self.tool_execute(tool, tool_input, agent_session)
        logger.info("actions = %s",
                    "\n".join([action.json(ensure_ascii=False) for action in agent_session.current_message.actions]))

    def get_tool_desc(self):
        return self.get_tool_desc_template().format(tools=[
            {"panguFunction": self.tool_map[tool].get_pangu_function()} for tool in self.tool_map])

    def normalize_prompt_template(self, prompt_str: str) -> str:
        """
        根据实际模型配置的占位符替换prompt模板中的占位符
        :param prompt_str: 输入prompt
        :return: 更换占位符后prompt
        """
        llm_module_property = self.llm.get_llm_config().llm_module_config.llm_module_property
        if llm_module_property and llm_module_property.unify_tag_prefix and\
                llm_module_property.unify_tag_prefix != "[unused9]":
            return prompt_str.replace("[unused9]", llm_module_property.unify_tag_prefix)\
                .replace("[unused10]", llm_module_property.unify_tag_suffix) \
                .replace("[unused11]", llm_module_property.unify_tool_tag_prefix) \
                .replace("[unused12]", llm_module_property.unify_tool_tag_suffix)
        return prompt_str

    def is_n2_module(self) -> bool:
        module_version = self.llm.get_llm_config().llm_module_config.module_version
        return module_version.startswith("38B") or module_version.startswith(self.MODULE_VERSION_PREFIX_N2)

    def is_plugin_v1_version(self, template_version) -> bool:
        return template_version not in [self.TEMPLATE_VERSION_AGENT_V2, self.TEMPLATE_VERSION_UNIFY]

    def get_template_version(self) -> str:
        module_version = self.llm.get_llm_config().llm_module_config.module_version
        return self.sub_str_after(module_version, "_")

    def get_react_template(self) -> PromptTemplate:
        template_version = self.get_template_version()
        if template_version == self.TEMPLATE_VERSION_AGENT_V2:
            return PromptTemplates.get("agent_react_pangu_2")
        elif template_version == self.TEMPLATE_VERSION_UNIFY:
            return PromptTemplates.get("agent_react_pangu_unify")
        else:
            return PromptTemplates.get("agent_react_pangu")

    def get_tool_desc_template(self) -> PromptTemplate:
        template_version = self.get_template_version()
        if template_version == self.TEMPLATE_VERSION_AGENT_V2 or template_version == self.TEMPLATE_VERSION_UNIFY:
            return PromptTemplates.get("agent_tool_desc_pangu_2")
        else:
            return PromptTemplates.get("agent_tool_desc_pangu")

    def set_cot_desc(self, cot_desc: str):
        self.llm.get_llm_config().llm_module_config.cot_desc = cot_desc



import json
import logging
from json import JSONDecodeError
from typing import Optional, List, Any, Iterator, AsyncIterator

import aiohttp
import requests
import sseclient
from langchain_core.callbacks import (
    CallbackManagerForLLMRun,
    AsyncCallbackManagerForLLMRun,
)
from langchain_core.language_models import LLM
from langchain_core.outputs import GenerationChunk
from requests.exceptions import ChunkedEncodingError

from langchain_pangu.pangukitsappdev.api.common_config import AUTH_TOKEN_HEADER
from langchain_pangu.pangukitsappdev.api.llms.base import get_llm_params
from langchain_pangu.pangukitsappdev.api.llms.llm_config import LLMConfig
from langchain_pangu.pangukitsappdev.auth.iam import IAMTokenProviderFactory, IAMTokenProvider


class PanGuLLM(LLM):
    temperature: Optional[float]
    max_tokens: Optional[int]
    top_p: Optional[float]
    presence_penalty: Optional[float]
    llm_config: LLMConfig
    streaming: Optional[bool]
    proxies: dict = {}
    pangu_url: Optional[str]
    token_getter: Optional[IAMTokenProvider]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pangu_url: str = self.llm_config.llm_module_config.url
        self.token_getter = IAMTokenProviderFactory.create(self.llm_config.iam_config)

    def _request_body(self, prompt: str, stream=True):
        rsp = {
            "prompt": prompt,
            "stream": stream,
            **get_llm_params(
                {
                    "temperature": self.temperature,
                    "max_tokens": self.max_tokens,
                    "top_p": self.top_p,
                    "presence_penalty": self.presence_penalty,
                }
            ),
        }
        return rsp

    def _headers(self):
        token = self.token_getter.get_valid_token()
        headers = (
            {AUTH_TOKEN_HEADER: token, "X-Agent": "pangu-kits-app-dev"}
            if token
            else {"X-Agent": "pangu-kits-app-dev"}
        )
        return headers

    def _call(
            self,
            prompt: str,
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> str:
        rsp = requests.post(
            self.pangu_url + "/text/completions",
            headers=self._headers(),
            json=self._request_body(prompt, stream=False),
            verify=False,
            stream=False,
            proxies=self.proxies,
        )

        if 200 == rsp.status_code:
            llm_output = rsp.json()
            text = llm_output["choices"][0]["text"]
        else:
            raise ValueError(
                "Call pangu llm failed, http status: %d, error response: %s",
                rsp.status_code,
                rsp.content,
            )

        return text

    async def _acall(
            self,
            prompt: str,
            stop: Optional[List[str]] = None,
            run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> str:
        proto = self.pangu_url.split("://")[0]
        async with aiohttp.ClientSession() as session:
            async with session.post(
                    self.pangu_url + "/text/completions",
                    headers=self._headers(),
                    json=self._request_body(prompt, stream=False),
                    verify_ssl=False,
                    proxy=self.proxies[proto] if proto in self.proxies else None,
            ) as rsp:
                if rsp.status == 200:
                    llm_output = await rsp.json()
                    text = llm_output["choices"][0]["text"]
                else:
                    raise ValueError(
                        "Call pangu llm failed, http status: %d",
                        rsp.status,
                    )
        return text

    def _stream(
            self,
            prompt: str,
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> Iterator[GenerationChunk]:
        rsp = requests.post(
            self.pangu_url + "/text/completions",
            headers=self._headers(),
            json=self._request_body(prompt),
            verify=False,
            stream=True,
            proxies=self.proxies,
        )
        try:
            rsp.raise_for_status()
            stream_client: sseclient.SSEClient = sseclient.SSEClient(rsp)
            for event in stream_client.events():
                # 解析出Token数据
                data_json = json.loads(event.data)
                if data_json.get("choices") is None:
                    raise ValueError(
                        f"Meet json decode error: {str(data_json)}, not get choices"
                    )
                chunk = GenerationChunk(text=data_json["choices"][0]["text"])
                yield chunk
                if run_manager:
                    run_manager.on_llm_new_token(chunk.text, chunk=chunk)
        except JSONDecodeError as ex:
            # [DONE]表示stream结束了
            pass
        except ChunkedEncodingError as ex:
            logging.warning(f"Meet error: %s", str(ex))

    async def _astream(
            self,
            prompt: str,
            stop: Optional[List[str]] = None,
            run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> AsyncIterator[GenerationChunk]:
        proto = self.pangu_url.split("://")[0]
        async with aiohttp.ClientSession() as session:
            async with session.post(
                    self.pangu_url + "/text/completions",
                    headers=self._headers(),
                    json=self._request_body(prompt),
                    verify_ssl=False,
                    proxy=self.proxies[proto] if proto in self.proxies else None,
            ) as rsp:
                while not rsp.closed:
                    line = await rsp.content.readline()
                    if line.startswith(b"data:[DONE]"):
                        rsp.close()
                        break
                    if line.startswith(b"data:"):
                        data_json = json.loads(line[5:])
                        if data_json.get("choices") is None:
                            raise ValueError(
                                f"Meet json decode error: {str(data_json)}, not get choices"
                            )
                        chunk = GenerationChunk(text=data_json["choices"][0]["text"])
                        yield chunk
                        if run_manager:
                            await run_manager.on_llm_new_token(chunk.text, chunk=chunk)
                    if line.startswith(b"event:"):
                        pass

    @property
    def _llm_type(self) -> str:
        return "pangu_llm"



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import logging
from typing import List

from langchain.prompts import PromptTemplate

from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi, ConversationMessage
from langchain_pangu.pangukitsappdev.api.skill.base import SimpleSkill
from langchain_pangu.pangukitsappdev.prompt.prompt_tmpl import PromptTemplates

logger = logging.getLogger(__name__)


class ConversationRewriteSkill:

    def __init__(self, llm: LLMApi):
        self.llm = llm
        self.prompt = PromptTemplates.get("skill_conversation_rewrite")

    def set_prompt(self, prompt_template: str):
        self.prompt = PromptTemplate.from_template(prompt_template, template_format="jinja2")

    def rewrite(self, messages: List[ConversationMessage]) -> str:
        query_write = SimpleSkill(self.prompt, self.llm)
        rewrite_result = query_write.execute({"messages": messages})
        logger.info("ConversationRewrite result: %s", rewrite_result)
        return rewrite_result



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
import logging
import os
import tempfile
from typing import Optional, List, Union

from langchain.docstore.document import Document
from langchain.document_loaders import UnstructuredFileLoader
from langchain.document_loaders.base import BaseLoader
from langchain.document_loaders.directory import FILE_LOADER_TYPE
from obs import LogConf, PutObjectHeader
from pydantic.v1 import Field

logger = logging.getLogger(__name__)

supported_files = [".doc", ".docx", ".pdf", ".txt", ".epub", ".eml", ".msg", ".md", ".pptx", ".html"]


class ObsLoader(BaseLoader):
    def load(self) -> List[Document]:
        # OBS本身是没有文件夹的概念的，文件夹实际上是创建了一个大小为0且对象名以“/”结尾的对象，这类对象与其他对象无任何差异，可以进行下载、删除等操作
        return self._load_from_object_paths(paths=self.object_paths)

    obs_server: str = Field(..., env="OBS_SERVER")
    access_key_id: str = Field(..., env="ACCESS_KEY_ID")
    secret_access_key: str = Field(..., env="SECRET_ACCESS_KEY")
    bucket_name: str
    object_paths: Optional[List[str]] = None
    loader_cls: Optional[FILE_LOADER_TYPE] = UnstructuredFileLoader,
    loader_kwargs: Optional[Union[dict, None]] = None,
    log_path: Optional[str] = None

    def _load_from_object_paths(self, paths: List[str]) -> List[Document]:
        """
        加载obs路径所有支持的文档内容，返回文档列表

        Args:
            paths: obs路径列表

        Returns:
            List[Document]: A list of Document objects representing
            the loaded documents.
        """
        docs = []

        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = f"{temp_dir}"
            os.makedirs(os.path.dirname(file_path), exist_ok=True)

            # 下载对象-文件下载
            for path in paths:
                filename, file_extension = os.path.splitext(path)
                if file_extension.lower() not in supported_files:
                    self._self_defined_load(docs, file_path)
                    break
                file_path = f"{temp_dir}/{filename}{file_extension}"
                try:
                    error_info = 'download file from obs failed.'
                    resp = self.client.getObject(self.bucket_name, path, downloadPath=file_path)
                    if resp.status < 300:
                        logger.info('requestId:', resp.requestId)
                        logger.info('url:', resp.body.url)
                    else:
                        logger.error('errorCode:', resp.errorCode)
                        logger.error('errorMessage:', resp.errorMessage)
                        break
                    error_info = 'load file type failed.'
                    try:
                        import unstructured  # noqa:F401
                    except ImportError:
                        raise ValueError(
                            "unstructured package not found, please install it with "
                            "`pip install unstructured`"
                        )
                    sub_docs = UnstructuredFileLoader(file_path).load()
                    docs.extend(sub_docs)
                except:
                    logger.error(error_info)
                    self._self_defined_load(docs, file_path)
        return docs

    def _self_defined_load(self, docs, file_path):
        error_info = 'file type not supported.try self-defined loader.'
        logger.error(error_info)
        try:
            sub_docs = self.loader_cls(file_path).load()
            docs.extend(sub_docs)
        except:
            error_info = 'self-defined loader failed.'
            logger.error(error_info)

    def __init__(
            self,
            location: str,
            access_key_id: str,
            secret_access_key: str,
            bucket_name: str,
            object_paths: Optional[List[str]] = None,
            obs_server: Optional[str] = None,
            loader_cls: Optional[FILE_LOADER_TYPE] = UnstructuredFileLoader,
            loader_kwargs: Optional[Union[dict, None]] = None,
            log_path: Optional[str] = None
    ):
        """Initialize with necessary components."""
        try:
            import obs
        except ImportError:
            raise ImportError(
                "Could not import obs python package. "
                "Please install it with `pip install esdk-obs-python`."
            )
        if obs_server:
            self.obs_server = obs_server
        else:
            self.obs_server = f"https://obs.{location}.myhuaweicloud.com"
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.location = location
        self.bucket_name = bucket_name
        self.object_paths = object_paths
        if loader_cls:
            self.loader_cls = loader_cls
        try:
            # 创建ObsClient实例
            self.client = obs.ObsClient(
                access_key_id=access_key_id,
                secret_access_key=secret_access_key,
                server=obs_server
            )
            # 日志初始化
            if log_path:
                # 指定日志配置文件路径，初始化ObsClient日志
                self.client.initLog(LogConf(log_path), obs_server + '-obs.log')

            # 桶初始化
            resp = self.client.headBucket(bucket_name)
            if resp.status == 404:
                logger.info('Bucket does not exist')
                # 创建桶 使用除华北-北京一（cn-north-1）以外的其他终端节点（endpoint）创桶时，必须指
                # 定Location，且Location必须与endpoint中的区域一致。桶会创建在location指定的区域。
                resp = self.client.createBucket(bucketName=bucket_name, location=location)
                if resp.status < 300:
                    logger.info('requestId:', resp.requestId)
                else:
                    logger.error('errorCode:', resp.errorCode)
                    logger.error('errorMessage:', resp.errorMessage)

            # 创建桶客户端
            self.bucket_client = self.client.bucketClient(bucket_name)
        except ValueError as e:
            raise ValueError(
                f"Your obs client string is mis-formatted. Got error: {e} "
            )

    def upload_object(self, local_paths: List[str], des_path: Optional[str] = None):
        headers = PutObjectHeader()
        headers.contentType = 'text/plain'
        # 上传本地文件/文件夹到指定桶中，不能超过5g，否则分段上传
        for local_path in local_paths:
            file_path, filename = os.path.split(local_path)
            if des_path:
                file_path = f"{des_path}/{filename}"
            try:
                resp = self.client.putFile(self.bucket_name, file_path, local_path, headers=headers)
                if resp.status < 300:
                    logger.info('requestId:', resp.requestId)
                    logger.info('etag:', resp.body.etag)
                    logger.info('versionId:', resp.body.versionId)
                    logger.info('storageClass:', resp.body.storageClass)
                else:
                    logger.error('errorCode:', resp.errorCode)
                    logger.error('errorMessage:', resp.errorMessage)
            except:
                logger.error("upload file to obs failed.")

    def delete_object(self, path: str):
        try:
            resp = self.client.deleteObject(self.bucket_name, path)

            if resp.status < 300:
                logger.info('requestId:', resp.requestId)
                logger.info('deleteMarker:', resp.body.deleteMarker)
                logger.info('versionId:', resp.body.versionId)
            else:
                logger.error('errorCode:', resp.errorCode)
                logger.error('errorMessage:', resp.errorMessage)
        except:
            logger.error("delete file on obs failed.")

    def close(self):
        # 关闭obsClient
        self.client.close()



from importlib import metadata

from langchain_pangu.chat_models import ChatPanGu
from langchain_pangu.embeddings import PanGuEmbeddings
from langchain_pangu.llms import PanGuLLM

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    # Case where package metadata is not available.
    __version__ = ""
del metadata  # optional, avoids polluting the results of dir(__package__)

__all__ = [
    "ChatPanGu",
    "PanGuEmbeddings",
    "PanGuLLM",
    "__version__",
]



#  Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.

from langchain_pangu.pangukitsappdev.api.llms.base import LLMApi
from langchain_pangu.pangukitsappdev.prompt.prompt_tmpl import PromptTemplates as pt
from langchain_pangu.pangukitsappdev.skill.doc.base import DocStuffSkill, DocRefineSkill, DocMapReduceSkill


class DocAskStuffSkill(DocStuffSkill):
    """文档问答stuff策略

    """

    def __init__(self, llm_api: LLMApi):
        super().__init__(pt.get("documents_stuff"), llm_api)


class DocAskRefineSkill(DocRefineSkill):
    """文档问答refine策略"""

    def __init__(self, llm_api: LLMApi):
        super().__init__(pt.get("documents_refine_qa"), pt.get("documents_refine_combine"), llm_api)


class DocAskMapReduceSkill(DocMapReduceSkill):
    """文档问答refine策略"""

    def __init__(self, llm_api: LLMApi, reduce_max_token: int = 2000):
        super().__init__(pt.get("documents_mapreduce_map"),
                         pt.get("documents_mapreduce_reduce"), llm_api, reduce_max_token)
